package nxm.sys.libm;

import nxm.sys.inc.Constants;

/**
   Performs the Mux operation on arrays. <br>
   See {@link Operator} for a description of the available I/O Modes.
   @author  Auto-Generated
   @see     Operator
*/
@SuppressWarnings("cast")
public final class Mux extends Operator implements Constants {

  /** Create a new instance. */
  public Mux() {
    super();
  }

  /** Create a new instance.
      @param flags The processing flags to use.
   */
  public Mux (int flags) {
    super(flags);
  }

  /** Create a new instance.
      @param scale The scale factor to use.
   */
  public Mux (double scale) {
    super(scale);
  }

  /** Create a new instance.
      @param flags The processing flags to use.
      @param scale The scale factor to use.
   */
  public Mux (int flags, double scale) {
    super(flags,scale);
  }

  public final int validate (char type, int n, int mode) {
    switch (mode) {
      case SSC: return n;
      default: return NOSUPPORT;
    }
  }

  public final int process (double[][] v, int n, int mode) {
    switch (mode) {
      case SSC: SSC(v[0],v[1],v[2],n); return n;
      default:  return NOSUPPORT;
    }
  }

  public final int process (double[][] v, int n, int mode, double scale) {
    switch (mode) {
      case SSC: SSC(v[0],v[1],v[2],n); return n;
      default:  return NOSUPPORT;
    }
  }

  /** Apply this operator to mode {@link Operator#SSC} data.
      @param a     The first  operand.
      @param b     The second  operand.
      @param c     The third  operand.
      @param n     The number of elements to operate on.
      @see #process(double[][],int,int)
   */
  public static final void SSC (double[] a, double[] b, double[] c, int n) {
    if (b.length == 1) {
      for (int i=0,j=0; i<n; i++,j+=2) {
        c[j]=a[i]; c[j+1]=b[0];
      }
    } else if (a.length == 1) {
      for (int i=0,j=0; i<n; i++,j+=2) {
        c[j]=a[0]; c[j+1]=b[i];
      }
    } else {
      for (int i=0,j=0; i<n; i++,j+=2) {
        c[j]=a[i]; c[j+1]=b[i];
      }
    }
  }

  public final int process (float[][] v, int n, int mode) {
    switch (mode) {
      case SSC: SSC(v[0],v[1],v[2],n); return n;
      default:  return NOSUPPORT;
    }
  }

  public final int process (float[][] v, int n, int mode, double scale) {
    switch (mode) {
      case SSC: SSC(v[0],v[1],v[2],n); return n;
      default:  return NOSUPPORT;
    }
  }

  /** Apply this operator to mode {@link Operator#SSC} data.
      @param a     The first  operand.
      @param b     The second  operand.
      @param c     The third  operand.
      @param n     The number of elements to operate on.
      @see #process(float[][],int,int)
   */
  public static final void SSC (float[] a, float[] b, float[] c, int n) {
    if (b.length == 1) {
      for (int i=0,j=0; i<n; i++,j+=2) {
        c[j]=a[i]; c[j+1]=b[0];
      }
    } else if (a.length == 1) {
      for (int i=0,j=0; i<n; i++,j+=2) {
        c[j]=a[0]; c[j+1]=b[i];
      }
    } else {
      for (int i=0,j=0; i<n; i++,j+=2) {
        c[j]=a[i]; c[j+1]=b[i];
      }
    }
  }

  public final int process (int[][] v, int n, int mode) {
    switch (mode) {
      case SSC: SSC(v[0],v[1],v[2],n); return n;
      default:  return NOSUPPORT;
    }
  }

  public final int process (int[][] v, int n, int mode, double scale) {
    switch (mode) {
      case SSC: SSC(v[0],v[1],v[2],n); return n;
      default:  return NOSUPPORT;
    }
  }

  /** Apply this operator to mode {@link Operator#SSC} data.
      @param a     The first  operand.
      @param b     The second  operand.
      @param c     The third  operand.
      @param n     The number of elements to operate on.
      @see #process(int[][],int,int)
   */
  public static final void SSC (int[] a, int[] b, int[] c, int n) {
    if (b.length == 1) {
      for (int i=0,j=0; i<n; i++,j+=2) {
        c[j]=a[i]; c[j+1]=b[0];
      }
    } else if (a.length == 1) {
      for (int i=0,j=0; i<n; i++,j+=2) {
        c[j]=a[0]; c[j+1]=b[i];
      }
    } else {
      for (int i=0,j=0; i<n; i++,j+=2) {
        c[j]=a[i]; c[j+1]=b[i];
      }
    }
  }

  public int getInputs() { return 2; }
  public int getOutputs() { return 1; }
}
