#!/bin/bash
#####################EXPERIMENTAL PURPOSES ONLY#######################
#                                                                    #
# This script removes the source code and test cases associated with #
# commands removed from commands.cnf with !-                         #
#                                                                    #
# Environment Variables:                                             #
#   NMROOT - Current NeXtMidas distribution                          #
#                                                                    #
# Author: Ryan McCall                                                #
# Since:  NeXtMidas 4.0.0                                            #
#                                                                    #
######################################################################

cmdsCnf=$NMROOT/nxm/sys/cfg/commands.cnf
testDir=$NMROOT/nxm/sys/test
primDir=$NMROOT/nxm/sys/prim
intrDir=$NMROOT/nxm/sys/intr
mcrDir=$NMROOT/nxm/sys/mcr
expDir=$NMROOT/nxm/sys/exp

# grep:
#     ^ = Start of line
#     !- = Regex to search for (! may need to be escaped)
# sed:
#     s/regex/replace/options
#     -e match following regular expression (can use multiple)
#     ^!\- = Start of line regex - same as grep\
#     \(...\) = Add contents to stack
#     [^\s]* = Match consecutive non-whitespace characters
#     \s.* = Match 1 whitespace character and remainder of the line
#     ^!\-\([^\s]*\)\s.* = Match start of line followed by !- and then followed by any number of non-whitespace
#                         characters, which are pushed to a stack, and then followed by a whitespace and any
#                                         other characters
#     \1 = Pop 1st element in regex stack
#     s/\*// = Replace all '*' characters with ''
# tr:
#     'A-Z' match uppercase
#     'a-z' replace with lowercase
#
# Note: Each each removed command is resulting in two entries in the array (command name portion and rest of line)
# Until this is addressed, this script uses 'index' to skip over every second entry

cmdList=$(grep '^!-' $cmdsCnf | sed -e 's/^!\-\([^\s]*\)\s.*/\1/' -e 's/\*//g' | tr 'A-Z' 'a-z')

declare -i index=0
for cmd in $(echo $cmdList) ; do
  if [ $index -eq 0 ] ; then
    capitalizedCMD=${cmd^}
    rm -f ${testDir}/test_${cmd}_*.m*
    rm -f ${primDir}/${cmd}.*
    rm -f ${primDir}/${cmd}\$*
    rm -f ${intrDir}/${capitalizedCMD}.*
    rm -f ${intrDir}/${capitalizedCMD}\$*
    rm -f ${mcrDir}/${cmd}.*
    rm -f ${expDir}/${cmd}.*
  fi
  index=($index+1)%2
done
